/*
 *
 ****************************************************************************
 *
 * MODULE:     r.out.xyz
 * AUTHOR(S):  Michael Shapiro
 *             Markus Neteler: added SURFER support
 *             Roger Miller added MODFLOW support and organization
 *             Kazukiyo Yamamoto: Modified for XYZ output
 * PURPOSE:    r.out.xyz: writes ASCII XYZ File
 * COPYRIGHT:  (C) 2000 by the GRASS Development Team
 *
 *             This program is free software under the GNU General Public
 *   	    	License (>=v2). Read the file COPYING that comes with GRASS
 *   	    	for details.
 *
 ****************************************************************************
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <grass/gis.h>
#include <grass/glocale.h>
#include <grass/gisdefs.h>

int write_XYZ(
int fd, 
FILE *fp, 
int nrows, 
int ncols, 
int out_type, 
int dp,
char answer, 
char *null_str)
{
	int row,col;
	void *ptr, *raster;
	char cell_buf[300];
	struct Cell_head region;
	DCELL x,y,zd;
	FCELL zf;

	G_get_window (&region);

	raster=G_allocate_raster_buf(out_type);

	for (row = 0; row < nrows; row++)
	{
		G_percent(row, nrows, 2);

		if (G_get_raster_row(fd, raster, row, out_type) < 0)
		   return(row);

		y = G_row_to_northing((double)row + 0.5, &region);
		for (col = 0, ptr = raster; col < ncols; col++, 
			   ptr = G_incr_void_ptr(ptr, G_raster_size(out_type))) 
		{
			x = G_col_to_easting((double)col + 0.5, &region);
			if(!G_is_null_value(ptr, out_type))
			{
			   if(out_type == CELL_TYPE)
			   fprintf (fp,"%f %f %d\n", x, y, *((CELL *) ptr));
	
			   else if(out_type == FCELL_TYPE)
			   {
			      zf=*((FCELL *) ptr);
			      sprintf(cell_buf, "%.*f", dp,zf);
			      fprintf (fp,"%f %f %s\n", x, y, cell_buf);
			   }
			   else if(out_type == DCELL_TYPE)
			   {
			      zd=*((DCELL *) ptr);
			      sprintf(cell_buf, "%.*f",  dp, zd);
			      fprintf (fp,"%f %f %s\n", x, y, cell_buf);
			   }
			}
			else{
                              if(!answer) 
                                  fprintf (fp,"%f %f %s\n", x, y, null_str);
                        }
		}
	}

	return (0);
}

int
main(int argc, char *argv[])
{
    RASTER_MAP_TYPE out_type, map_type;
    char *name; 
    char *mapset;
    char *null_str;
    int fd;
    int nrows, ncols, dp, width;
    int rc;
    FILE *fp;
    struct GModule *module;
    struct
    {
	struct Option *map ;
	struct Option *output ;
	struct Option *dp ;
        struct Option *width ;
	struct Option *null ;
    } parm;
    struct
    {
        struct Flag *non_null;
        struct Flag *int_out;
    } flag;

    G_gisinit(argv[0]);

    module = G_define_module();
    module->description =
		_("Converts a raster map layer into an ASCII text file.");

/* Define the different options */

    parm.map = G_define_option() ;
    parm.map->key        = "input";
    parm.map->type       = TYPE_STRING;
    parm.map->required   = YES;
    parm.map->gisprompt  = "old,cell,raster" ;
    parm.map->description= _("Name of an existing raster map") ;

    parm.output = G_define_option() ;
    parm.output->key        = "output";
    parm.output->type       = TYPE_STRING;
    parm.output->required   = NO;
    parm.output->gisprompt  = "any,cell,raster" ;
    parm.output->description= _("Name of an output ASCII XYZ ");

    parm.dp = G_define_option() ;
    parm.dp->key        = "dp";
    parm.dp->type       = TYPE_INTEGER;
    parm.dp->required   = NO;
    parm.dp->description= _("Number of significant digits (floating point only)") ;

    parm.width = G_define_option() ;
    parm.width->key        = "width" ;
    parm.width->type       = TYPE_INTEGER;
    parm.width->required   = NO;
    parm.width->description=_("Number of values printed before wrapping a line (only SURFER or MODFLOW format)") ;

    parm.null = G_define_option() ;
    parm.null->key        = "null";
    parm.null->type       = TYPE_STRING;
    parm.null->required   = NO;
    parm.null->answer     = "*";
    parm.null->description= _("String to represent null cell (GRASS grid only)") ;

    flag.non_null = G_define_flag();
    flag.non_null->key = 'n';
    flag.non_null->description = _("Suppress printing of null values");

    flag.int_out = G_define_flag();
    flag.int_out->key = 'i';
    flag.int_out->description = _("Force output of integer values");

    if (G_parser(argc, argv))
       	exit (-1);

    if(parm.dp->answer)
    {
        if(sscanf(parm.dp->answer,"%d",&dp)!=1)
           G_fatal_error("failed to interpret dp as an integer\n");
        if(dp>20 || dp < 0)
           G_fatal_error("dp has to be from 0 to 20\n");
    }

    width=10;
    if(parm.width->answer)
    {
        if(sscanf(parm.width->answer,"%d",&width)!=1)
            G_fatal_error("failed to interpret width as an integer\n");
    }

    null_str = parm.null->answer;

    name = parm.map->answer;
    mapset = G_find_cell2 (name, "");

    map_type = G_raster_map_type(name, mapset);
    if(!flag.int_out->answer)
	 out_type = map_type;
    else out_type = CELL_TYPE;

    if(!parm.dp->answer)
    {
        dp=6;
        if(out_type==DCELL_TYPE) dp=16;
    }

    if (mapset == NULL)
    {
        G_fatal_error ("Cell file [%s] not found\n",name);
    }

    nrows = G_window_rows();
    ncols = G_window_cols();

/* open ascii file for writing or use stdout */
    if(parm.output->answer && strcmp("-",parm.output->answer)!=0)
    {
        if(NULL == (fp = fopen(parm.output->answer, "w")))
           G_fatal_error("Not able to open file [%s]", parm.output->answer);
    }
    else
       fp = stdout;

/* open raster file */
    fd = G_open_cell_old (name, mapset);
    if (fd < 0)
        exit(1);

/* process the requested output format */
    rc=write_XYZ(fd,fp,nrows,ncols,out_type,dp,flag.non_null->answer,null_str);
    if(rc)
    {
        G_fatal_error("Read failed at row [%d]\n",rc);
    }
  
/* tidy up and go away */
    G_close_cell(fd);
    fclose(fp);
    exit(0);
}
